"""Base class for serving predictions."""
from flask import Flask, request, jsonify
from flask_restful import Resource, Api
import numpy as np
from meinheld import server, middleware

from .utils import make_serializable
from .log_utils import get_logger

logger = get_logger(__name__)


class PredictionServer(object):
    """Easy deploy class."""

    def __init__(self, predict, input_validation=lambda data: (True, None)):
        """Initialize class with prediction function.

        Arguments:
            - predict (fn): function that takes a numpy array of features as input,
                and returns a prediction of targets
            - input_validation (fn): takes a numpy array as input;
                returns True if validation passes and False otherwise
        """
        self.predict = predict
        self.app = Flask('{}_{}'.format(self.__class__.__name__, type(predict).__name__))
        self.api = Api(self.app, catch_all_404s=True)
        self._create_prediction_endpoint(input_validation)
        logger.info('Model predictions now being served from endpoint /predictions (available via POST)')
        self.app.logger.setLevel(logger.level)  # TODO: separate configuration for API loglevel

    def __repr__(self):
        """String representation."""
        return '<PredictionsServer: {}>'.format(type(self.predict).__name__)

    def _create_prediction_endpoint(self, input_validation=lambda data: (True, None)):
        """Create an endpoint to serve predictions.

        Arguments:
            - input_validation (fn): takes a numpy array as input;
                returns True if validation passes and False otherwise
        """
        # copy instance variables to local scope for resource class
        predict = self.predict
        logger = self.app.logger

        # create restful resource
        class Predictions(Resource):
            @staticmethod
            def post():
                # parse request data
                data = request.get_json()
                logger.debug('Received JSON data of length {:,}'.format(len(data)))

                # convert to numpy array
                data = np.array(data)
                logger.debug('Converted JSON data to Numpy array with shape {}'.format(data.shape))

                # sanity check using user defined callback (default is no check)
                validation_pass, validation_reason = input_validation(data)
                if not validation_pass:
                    # if validation fails, log the reason code, log the data, and send a 400 response
                    validation_message = 'Input validation failed with reason: {}'.format(validation_reason)
                    logger.error(validation_message)
                    logger.debug('Data: {}'.format(data))
                    response = jsonify(dict(message=validation_message))
                    response.status_code = 400
                    return response

                try:
                    prediction = predict(data)
                except Exception as e:
                    # log exception and return the message in a 500 response
                    logger.error('{} exception: {}'.format(type(e).__name__, e))
                    response = jsonify(dict(message=str(e)))
                    response.status_code = 500
                    return response
                logger.debug('Predictions generated with shape {}'.format(prediction.shape))
                return prediction.tolist()

        # map resource to endpoint
        self.api.add_resource(Predictions, '/predictions')

    def create_info_endpoint(self, name, data):
        """Create an endpoint to serve info GET requests."""
        # make sure data is serializable
        data = make_serializable(data)

        # create generic restful resource to serve static JSON data
        class InfoBase(Resource):
            @staticmethod
            def get():
                return data

        def info_factory(name):
            """Return an Info derivative resource."""
            class NewClass(InfoBase):
                pass
            NewClass.__name__ = "{}_{}".format(name, InfoBase.__name__)
            return NewClass

        path = '/info/{}'.format(name)
        self.api.add_resource(info_factory(name), path)
        logger.info('Added informational resource at path {} (available via GET)'.format(path))
        logger.debug('Endpoint {} will now serve the following static data:\n{}'.format(path, data))

    def serve(self, host='127.0.0.1', port=5000):
        """Serve predictions as an API endpoint."""
        # self.app.run(host=host, port=port)
        server.listen((host, port))
        server.run(middleware.WebSocketMiddleware(self.app))



"""Test SklearnServer."""
import unittest
import json
from sklearn.datasets import load_iris, load_boston

from serveit.sklearn_server import SklearnServer
from tests.test_prediction_server import PredictionServerTest


class SklearnServerTest(PredictionServerTest, object):
    """Base class to test the Scikit-Learn server.

    SklearnServerTest should be inherited by a class that has a sklearn `model`
    attribute, and calls `SklearnServerTest._setup()` after instantiation.
    That class should also inherit from `unittest.TestCase` to ensure tests are executed.
    """

    def _setup(self, data):
        """Set up method to be called before each unit test."""
        super(SklearnServerTest, self)._setup(self.model.fit, data)
        self.sklearn_server = SklearnServer(self.model, self.model.predict)
        self.app = self.sklearn_server.app.test_client()

    def test_model_info_none(self):
        """Verify 404 response if /info/model endpoint not yet created."""
        response = self.app.get('/info/model')
        self.assertEqual(response.status_code, 404)

    def test_model_info(self):
        """Test model info endpoint."""
        self.sklearn_server.create_model_info_endpoint()
        app = self.sklearn_server.app.test_client()
        response = app.get('/info/model')
        response_data = json.loads(response.get_data())
        self.assertGreater(len(response_data), 3)
        expected_keys = ['classes_']
        for key in expected_keys:
            self.assertIn(key, response_data)

    def test_input_validation(self):
        """Add simple input validator and make sure it triggers."""
        import numpy as np

        # model input validator
        def feature_count_check(data):
            # check num dims
            if data.ndim != 2:
                return False, 'Data should have two dimensions.'
            # check number of columns
            if data.shape[1] != self.data.data.shape[1]:
                reason = '{} features required, {} features provided'.format(
                    data.shape[1], self.data.data.shape[1])
                return False, reason
            # validation passed
            return True, None

        # set up test server
        sklearn_server = SklearnServer(self.model, self.model.predict, feature_count_check)
        app = sklearn_server.app.test_client()

        # generate sample data
        sample_idx = np.random.randint(self.data.data.shape[0], size=100)
        sample_data = self.data.data[sample_idx, :]

        # post good data, verify 200 response
        response = app.post(
            '/predictions',
            headers={'Content-Type': 'application/json'},
            data=json.dumps(sample_data.tolist()),
        )
        self.assertEqual(response.status_code, 200)

        # post bad data (drop a single column), verify 400 response
        response = app.post(
            '/predictions',
            headers={'Content-Type': 'application/json'},
            data=json.dumps(sample_data[:, :-1].tolist()),
        )
        self.assertEqual(response.status_code, 400)
        response_data = json.loads(response.get_data())
        expected_reason = '{} features required, {} features provided'.format(
            self.data.data.shape[1] - 1, self.data.data.shape[1])
        self.assertIn(expected_reason, response_data['message'])


class IrisLogisticRegressionTest(unittest.TestCase, SklearnServerTest):
    """Test SklearnServer with LogisticRegression fitted on iris data."""

    def setUp(self):
        """Unittest set up."""
        from sklearn.linear_model import LogisticRegression
        self.model = LogisticRegression()
        super(IrisLogisticRegressionTest, self)._setup(load_iris())


class IrisSvcTest(unittest.TestCase, SklearnServerTest):
    """Test SklearnServer with SVC fitted on iris data."""

    def setUp(self):
        """Unittest set up."""
        from sklearn.svm import SVC
        self.model = SVC()
        super(IrisSvcTest, self)._setup(load_iris())


class IrisRandomForestTest(unittest.TestCase, SklearnServerTest):
    """Test SklearnServer with RandomForestClassifier fitted on iris data."""

    def setUp(self):
        """Unittest set up."""
        from sklearn.ensemble import RandomForestClassifier
        self.model = RandomForestClassifier()
        super(IrisRandomForestTest, self)._setup(load_iris())


class BostonLinearRegressionTest(unittest.TestCase, PredictionServerTest):
    """Test PredictionServer with LogisticRegression fitted on housing data."""

    def setUp(self):
        """Unittest set up."""
        from sklearn.linear_model import LinearRegression
        self.model = LinearRegression()
        super(BostonLinearRegressionTest, self)._setup(self.model.fit, load_boston())


class BostonSvrTest(unittest.TestCase, PredictionServerTest):
    """Test PredictionServer with SVR fitted on housing data."""

    def setUp(self):
        """Unittest set up."""
        from sklearn.svm import SVR
        self.model = SVR()
        super(BostonSvrTest, self)._setup(self.model.fit, load_boston())


class BostonRfrTest(unittest.TestCase, PredictionServerTest):
    """Test PredictionServer with LogisticRegression fitted on housing data."""

    def setUp(self):
        """Unittest set up."""
        from sklearn.ensemble import RandomForestRegressor
        self.model = RandomForestRegressor()
        super(BostonRfrTest, self)._setup(self.model.fit, load_boston())

if __name__ == '__main__':
    unittest.main()



"""SkLearn model serving."""
from flask_restful import Resource

from .prediction_server import PredictionServer
from .utils import make_serializable


class SklearnServer(PredictionServer):
    """Deploys any Scikit-Learn model prediction method to an API endpoint.

    Also provides methods to support the creation of new endpoints to
    provide information about the model, expected features, and targets.
    """

    def __init__(self, model, *args, **kwargs):
        """Initialize class with model and prediction function."""
        super(SklearnServer, self).__init__(*args, **kwargs)
        self.model = model

    def __repr__(self):
        """String representation."""
        return '<EasyDeploySklearn: {}>'.format(self.model)

    def create_model_info_endpoint(self, path='/info/model'):
        """Create an endpoint to serve info GET requests."""
        model = self.model

        # parse model details
        model_details = {}
        for key, value in model.__dict__.items():
            model_details[key] = make_serializable(value)

        # create generic restful resource to serve model information as JSON
        class ModelInfo(Resource):
            @staticmethod
            def get():
                return model_details

        self.api.add_resource(ModelInfo, path)
        self.app.logger.info('Added informational resource at path {} (available via GET)'.format(path))
        self.app.logger.debug('Endpoint {} will now serve the following static data:\n{}'.format(path, model_details))


if __name__ == '__main__':
    from sklearn.datasets import load_iris
    from sklearn.linear_model import LogisticRegression

    # fit a model on the Iris dataset
    data = load_iris()
    reg = LogisticRegression()
    reg.fit(data.data, data.target)

    # deploy model to a SkLearnServer
    eds = SklearnServer(reg, reg.predict)

    # add informational endpoints
    eds.create_model_info_endpoint()
    eds.create_info_endpoint('features', data.feature_names)
    eds.create_info_endpoint('target_labels', data.target_names.tolist())

    # start API
    eds.serve()



"""Utility methods."""
import json


def make_serializable(data):
    """Ensure data is serializable."""
    if is_serializable(data):
        return data

    # if numpy array convert to list
    try:
        return data.tolist()
    except AttributeError:
        pass

    # try serializing each child element
    if isinstance(data, dict):
        return {key: make_serializable(value) for key, value in data.items()}
    if hasattr(data, '__iter__'):
        return [make_serializable(element) for element in data]

    # last resort: convert to string
    return str(data)


def is_serializable(data):
    """Check if data is serializable."""
    try:
        json.dumps(data)
        return True
    except TypeError:
        return False
